#!python
# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

import os
import sys
import time
import re
import logging

import p4lib
import uriparse
import koprocessutils
from koAsyncOperationUtils import koAsyncOperationBase

from xpcom import components, nsError, ServerException, COMException
from xpcom.server import WrapObject, UnwrapObject


#---- support routines

def _unique(s):
    """Return a list of the elements in s, in arbitrary order, but without
    duplicates. (_Part_ of the Python Cookbook recipe.)
    """
    # get the special case of an empty s out of the way, very rapidly
    n = len(s)
    if n == 0:
        return []

    # Try using a dict first, as that's the fastest and will usually work.
    u = {}
    try:
        for x in s:
            u[x] = 1
    except TypeError:
        del u  # move on to the next method
    else:
        return u.keys()

def _splitall(path):
    """Split the given path into all its directory parts and return the list
    of those parts (see Python Cookbook recipe for test suite.
    """
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts

def _commonprefix(paths):
    """An os.path.commonprefix() more suited to actual paths.
    
    It returns the common path prefix of the given paths or None, if there
    is no such common prefix.
    """
    if not paths:
        return None
    splitpaths = [_splitall(path) for path in paths]
    commonprefix = []
    for set in zip(*splitpaths):
        # This path element is part of the common prefix if it is the same
        # for every give path.
        elem = set[0]
        if sys.platform.startswith("win"):
            # Case-insensitive comparison on Windows.
            set = [p.lower() for p in set]
        if len(_unique(set)) == 1:
            commonprefix.append(elem)
        else:
            break
    if commonprefix:
        retval = os.path.join(*commonprefix)
    else:
        retval = None
    #print "_commonprefix(%r) == %r" % (paths, retval)
    return retval



#---- component implementation

class KoP4:
    # Satisfy koISCC.name
    name = "p4"
    # XPCOM component registration settings.
    _com_interfaces_ = [components.interfaces.koISCC, components.interfaces.nsIObserver]
    _reg_desc_ = "Komodo P4 Support"
    _reg_contractid_ = "@activestate.com/koSCC?type=" + name + ";1"
    _reg_clsid_ = "{7BD69117-A206-410b-A2C1-FAC34976734A}"
    _reg_categories_ = [
         ("category-komodo-scc", name),
         ]

    def __init__(self):
        # generic cvs instance to do work with
        self.p4 = p4lib.P4()
        self._observerSvc = components.classes["@mozilla.org/observer-service;1"].\
            getService(components.interfaces.nsIObserverService)
        self._fileStatusSvc = components.classes["@activestate.com/koFileStatusService;1"].\
                getService(components.interfaces.koIFileStatusService);
        self._globalPrefs = components.classes["@activestate.com/koPrefService;1"].\
            getService(components.interfaces.koIPrefService).prefs
        self._wrapped_self = WrapObject(self, components.interfaces.nsIObserver)
        self._globalPrefs.prefObserverService.addObserver(self._wrapped_self,
                                                          'p4Executable', 0)
        self._userPath = koprocessutils.getUserEnv()["PATH"].split(os.pathsep)
        self._encodingSvc = components.classes['@activestate.com/koEncodingServices;1'].\
                                getService(components.interfaces.koIEncodingServices)

        self.reasonNotFunctional = None
        self.isFunctional = None
        self._executable = None
        self._setExecutableFromPrefs()
        self.redetermineIfFunctional()
    
    def observe(self, subject, topic, value):
        if topic == 'p4Executable':
            self._setExecutableFromPrefs()
            
    def _setExecutableFromPrefs(self):
        executable = self._globalPrefs.getStringPref('p4Executable')
        if self._executable != executable:
            self.set_executable(executable)
            try:
                self._observerSvc.notifyObservers(self, 'subsystem_changed', 'p4')
            except COMException:
                pass # Ignore error if there is no observer.
        
    def set_executable(self, executable):
        if executable:
            self._executable = executable
            if os.path.exists(executable):
                self.p4.p4 = self._executable
        elif sys.platform.startswith("win"):
            self.p4.p4 = "p4.exe"
        else:
            self.p4.p4 = "p4"
        
    def get_executable(self):
        return self._executable
        
    def redetermineIfFunctional(self):
        import which
        assert which._version_ >= (1,0,0)
            
        if not self._executable:
            try:
                if sys.platform.startswith('win'):
                    exts = ['.exe']
                else:
                    exts = None
                p4 = which.which('p4', exts=exts, path=self._userPath)
            except which.WhichError:
                p4 = None
        elif os.path.exists(self._executable):
            p4 = self._executable
        else:
            p4 = None
        
        if p4 is None:
            self.isFunctional = 0
            self.reasonNotFunctional = "no 'p4' executable could be found"
        else:
            self.isFunctional = 1
            self.reasonNotFunctional = None
        try:
            self._observerSvc.notifyObservers(self, 'subsystem_changed', 'p4')
        except COMException:
            pass # Ignore error if there is no observer.

    def _splitFile(self, file):
        # figure out what repositories files are in
        # and split them up into seperate lists
        # we also convert to local paths here
        try:
            path = uriparse.URIToLocalPath(file);
            if path:
                if os.path.isdir(path):
                    return path,'./...'
                path = os.path.normpath(path)
                basedir = os.path.dirname(path)
                filename = os.path.basename(path)
                
                # make sure the path is in cvs
                #root = self.p4.have(path)
                #if root:
                #    return basedir, filename
                return basedir, filename
        except p4lib.P4LibError, e:
            # we'll ignore files that are not in perforce
            pass
        return None,None
    
    def _splitFiles(self, files):
        # figure out what repositories files are in
        # and split them up into seperate lists
        # we also convert to local paths here
        _files = {}
        for file in files:
            basedir,filename = self._splitFile(file)
            if basedir:
                if not basedir in _files:
                    _files[basedir] = []
                _files[basedir].append(filename)
        return _files
    
    def _convertFiles(self, files):
        _files = []
        for file in files:
            path = uriparse.URIToLocalPath(file)
            path = os.path.normpath(path)
            _files.append(path)
        return _files
    
#    diff -- Display diff of client file with depot file
#
#    p4 diff [ -d<flag> -f -sa -sd -se -sr -t ] [ file[rev] ... ]
#
#        Run diff (on the client) of a client file against the corresponding
#        revision in the depot. The file is only compared if the file is
#        opened for edit or the revision provided with the file argument is
#        not the same as the revision had by the client.  See 'p4 help
#        revisions' for help specifying revisions.
#
#        If no file argument is given, diff all open files.
#        This can be used to view pending changelists.
#
#        The -d<flag> passes a flag to the built-in diff routine to modify
#        the output: -dn (RCS), -dc (context), -ds (summary), -du (unified).
#
#        The -f flag forces a diff for every file, regardless of whether
#        they are opened or if the client has the named revision.
#        This can be used to verify the client contents.
#
#        The -s flag reduces the output of diff to the names of files
#        satisfying the following criteria:
#
#                -sa     Opened files that are different than the revision
#                        in the depot, or missing.
#
#                -sd     Unopened files that are missing on the client.
#
#                -se     Unopened files that are different than the revision
#                        in the depot.
#
#                -sr     Opened files that are the same as the revision in the
#                        depot.
#
#        The -t flag forces 'p4 diff' to diff even files with non-text
#        (binary) types.
#
#        If the environment variable $P4DIFF is set then the named program is
#        used rather than the implementation of diff included in the client.
#        The -d<flag>command can be used to pass arguments to the
#        external program.  The -s flag is only implemented internally.
    def _do_diff(self, files, options, external, p4=None):
        #print "p4 diff ",repr(files)
        _files = self._splitFiles(files)
        diff = ''
        if p4 is None:
            p4  = self.p4
        for basedir in _files.keys():
            result, raw = p4.diff(files=_files[basedir],
                                  diffFormat=options,
                                  _raw=1,
                                  cwd=basedir,
                                  p4diff=external)
            rawerror = raw.get('stderr', "")
            if rawerror:
                # There was some type of error
                raise p4lib.P4LibError(rawerror)

            if 'stdout' in raw:
                diff += raw['stdout']

        # Convert line-endings to the currently expected
        import eollib
        eolpref = self._globalPrefs.getStringPref('endOfLine')
        try:
            desiredEOL = eollib.eolPref2eol[eolpref]
        except KeyError:
            desiredEOL = eollib.EOL_PLATFORM
        diff = eollib.convertToEOLFormat(diff, desiredEOL)

        return diff

    def diff(self, files, options, external, async_callback):
        return runP4CommandAsynchronously("diff", self, self._do_diff,
                                          files, False, async_callback,
                                          files, options, external)

#    diff -- Display diff of client file with depot file
#
#    p4 diff [ -d<flag> -f -sa -sd -se -sr -t ] [ file[rev] ... ]
#
#        Run diff (on the client) of a client file against the corresponding
#        revision in the depot. The file is only compared if the file is
#        opened for edit or the revision provided with the file argument is
#        not the same as the revision had by the client.  See 'p4 help
#        revisions' for help specifying revisions.
#
#        If no file argument is given, diff all open files.
#        This can be used to view pending changelists.
#
#        The -d<flag> passes a flag to the built-in diff routine to modify
#        the output: -dn (RCS), -dc (context), -ds (summary), -du (unified).
#
#        The -f flag forces a diff for every file, regardless of whether
#        they are opened or if the client has the named revision.
#        This can be used to verify the client contents.
#
#        The -s flag reduces the output of diff to the names of files
#        satisfying the following criteria:
#
#                -sa     Opened files that are different than the revision
#                        in the depot, or missing.
#
#                -sd     Unopened files that are missing on the client.
#
#                -se     Unopened files that are different than the revision
#                        in the depot.
#
#                -sr     Opened files that are the same as the revision in the
#                        depot.
#
#        The -t flag forces 'p4 diff' to diff even files with non-text
#        (binary) types.
#
#        If the environment variable $P4DIFF is set then the named program is
#        used rather than the implementation of diff included in the client.
#        The -d<flag>command can be used to pass arguments to the
#        external program.  The -s flag is only implemented internally.
    def _do_diffRevisions(self, fileuri1, rev1, fileuri2, rev2, filepath,
                          options, external, p4=None):
        #print "p4 diffRevisions ", repr(fileuri1), rev1, repr(fileuri2), rev2
        basedir, filename = self._splitFile(filepath)
        #print "basedir",basedir
        if p4 is None:
            p4 = self.p4
        result, raw = self.p4.diffRevisions(fileuri1, rev1,
                                            fileuri2, rev2,
                                            diffFormat=options,
                                            _raw=1,
                                            cwd=basedir,
                                            p4diff=external)

        rawerror = raw.get('stderr', "")
        if rawerror:
            # There was some type of error
            #print "stderr:", rawerror
            raise p4lib.P4LibError(rawerror)

        if 'stdout' in raw:
            diff = raw['stdout']
        else:
            diff = result['text']
        # Convert line-endings to the currently expected
        import eollib
        eolpref = self._globalPrefs.getStringPref('endOfLine')
        try:
            desiredEOL = eollib.eolPref2eol[eolpref]
        except KeyError:
            desiredEOL = eollib.EOL_PLATFORM
        diff = eollib.convertToEOLFormat(diff, desiredEOL)
        # Try encoding it with komodo's unicode encoding service
        diff, encoding, bom = self._encodingSvc.getUnicodeEncodedString(diff)

        return diff

    def diffRevisions(self, fileuri1, rev1, fileuri2, rev2, filepath, options,
                      external, async_callback):
        return runP4CommandAsynchronously("diffRevisions", self, self._do_diffRevisions,
                                          [filepath], False, async_callback,
                                          fileuri1, rev1, fileuri2, rev2,
                                          filepath, options, external)

#    filelog -- List revision history of files
#
#    p4 filelog [ -i -l -L -t -m maxRevs ] file[rev] ...
#
#        List the revision history of the files named, working backwards
#        from the latest revision to the first.  If the file specification
#        includes a revision, the command limits its output to revisions at
#        or previous to the given revision
#
#        The -i flag follows branches.  If a file was created by branching,
#        'p4 filelog' also lists the revisions of the source file, but
#        only those revisions leading up to the branch point.
#
#        The -t flag displays the time as well as the date.
#
#        The -l flag produces long output with the full text of the
#        changelist descriptions.
#
#        The -L flag produces long output with the full text of the
#        changelist descriptions truncated to 250 characters.
#
#        The -m maxRevs displays at most 'maxRevs' revisions per file.
    def _do_history(self, fileuri, options, p4=None):
        #print "p4 filelog ", options, repr(fileuri)
        # ,cwd=basedir
        basedir, filename = self._splitFile(fileuri)
        longOutput = ('L' in options)
        followIntegrations = ('i' in options)
        if p4 is None:
            p4 = self.p4
        result, raw = p4.filelog(files=[filename],
                                 longOutput=longOutput,
                                 followIntegrations=followIntegrations,
                                 _raw=1,
                                 cwd=basedir)

        rawerror = raw.get('stderr', "")
        if rawerror:
            # There was some type of error
            raise p4lib.P4LibError(rawerror)

        #result:
        #Returns a list of hits. Each hit is a dict with the following
        #keys: 'depotFile', 'revs'. 'revs' is a list of dicts, each
        #representing one submitted revision of 'depotFile' and
        #containing the following keys: 'action', 'change', 'client',
        #'date', 'type', 'notes', 'rev', 'user', 'description'.

        # Now, split up the history rows

        koHistoryList = []
        #print result
        for filehit in result:
            url  = filehit['depotFile']
            revs = filehit['revs']
            for revision in revs:
                koHistory = components.classes["@activestate.com/koSCCHistoryItem;1"].\
                                createInstance(components.interfaces.koISCCHistoryItem);
                koHistory.version = str(revision['rev'])
                koHistory.change  = str(revision['change'])
                koHistory.author  = revision['user']
                koHistory.date    = revision['date']
                koHistory.action  = revision['action']
                koHistory.message = revision['description']
                koHistory.uri     = url
                koHistoryList.append(koHistory)
        return koHistoryList

    def history(self, fileuri, options, async_callback):
        return runP4CommandAsynchronously("history", self, self._do_history,
                                          [fileuri], False, async_callback,
                                          fileuri, options)

#    add -- Open a new file to add it to the depot
#
#    p4 add [ -c changelist# ] [ -t filetype ] file ...
#
#        Open a new file for adding to the depot.  If the file exists
#        on the client it is read to determine if it is text or binary.
#        If it does not exist it is assumed to be text.  The file must
#        either not exist in the depot, or it must be deleted at the
#        current head revision.  Files may be deleted and re-added.
#
#        If the -c flag is given the open files are associated with the
#        specified pending changelist number; otherwise the open files are
#        associated with the default changelist.
#
#        If file is already open it is moved into the specified pending
#        changelist.  It is not permissible to reopen a file for add unless
#        it was already open for add.
#
#        If -t filetype is given the file is explicitly opened as that
#        filetype.  Otherwise, the filetype is determined by the file
#        name-to-type mapping table managed by 'p4 typemap'.  If the file
#        name is not mapped in that table, 'p4 add' senses the filetype
#        by examining the file's contents and execution permission bits.
#        See 'p4 help filetypes' for a complete list.
    def _do_add(self, files, mode, message, p4=None):
        #add(self, files, mode=None, msg=None, cwd=None)
        # first we figure out what repositories files are in
        # and split them up into seperate lists
        # we init the cvs object with the root we retreived
        _files = self._splitFiles(files)
        if p4 is None:
            p4  = self.p4
        output = ''
        for basedir in _files.keys():
            result, raw = self.p4.add(files=_files[basedir],
                                      _raw=1, cwd=basedir)
            if 'stderr' in raw:
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
        # do notification that status may have changed
        # we have to use the original urls passed in since
        # they are modified to local paths in _splitFiles
        forceRefresh = False
        self._fileStatusSvc.updateStatusForUris(files, forceRefresh)
        return output

    def add(self, files, mode, message, async_callback=None):
        return runP4CommandAsynchronously("add", self, self._do_add,
                                          files, False, async_callback, files,
                                          mode, message)

#    submit -- Submit open files to the depot
#
#    p4 submit [ -s ]
#    p4 submit [ -s ] files
#    p4 submit -c changelist#
#    p4 submit -i [ -s ]
#
#        'p4 submit' commits a pending changelist and its files to the depot.
#
#        With no argument 'p4 submit' attempts to submit all files in the
#        'default' changelist.  Submit provides the user with a dialog
#        similar to 'p4 change' so the user can compose a changelist
#        description.  In this dialog the user is presented with the list
#        of files open in changelist 'default'.  Files may be deleted from
#        this list but they cannot be added.  (Use an open command (edit,
#        add, delete) to add additional files to a changelist.)
#
#        If a (single) file pattern is given, only those files in
#        the 'default' changelist that match the pattern will be submitted.
#
#        The -c flag submits the numbered pending changelist that has been
#        previously created with 'p4 change' or a failed 'p4 submit'.
#
#        The -i flag causes a changelist specification (including files to be
#        submitted) to be read from the standard input.  The user's editor
#        is not invoked.
#
#        The -s flag extends the list of jobs to include the fix status
#        for each job, which becomes the job's status when the changelist
#        is committed.  See 'p4 help change' for more notes on this option.
#
#        Before committing a changelist submit locks all associated files not
#        already locked.  If any file cannot be locked, or if the submit
#        fails for any other reason the files are left open in a newly
#        created pending changelist.
#
#        Submit is guaranteed to be atomic.  Either all files will be
#        updated in the depot as a unit or none will be.
    def _do_commit(self, files, message, options, p4=None):
        #print "p4 commit ",repr(files)
        # p4lib saves message to a message for file
        _split = self._splitFiles(files)
        _files = self._convertFiles(files)
        basedir = _commonprefix(_split.keys())
        if p4 is None:
            p4  = self.p4
        output = ''
        result, raw = p4.submit(_files,
                                description=message,
                                _raw=1,
                                cwd=basedir)

        raw_stderr = raw.get('stderr')
        if raw_stderr:
            # XXX - Should we be always raising an exception, or can some
            #       stderr warnings trickle through on a sucessful commit?
            raise p4lib.P4LibError(raw_stderr)
        if 'stdout' in raw:
            output += raw['stdout']
        
        forceRefresh = False
        self._fileStatusSvc.updateStatusForUris(files, forceRefresh)
        return output

    def commit(self, files, message, options, async_callback):
        return runP4CommandAsynchronously("commit", self, self._do_commit,
                                          files, True, async_callback, files,
                                          message, options)

#    delete -- Open an existing file to delete it from the depot
#
#    p4 delete [ -c changelist# ] file ...
#
#        Opens a file that currently exists in the depot for deletion.
#        If the file is present on the client it is removed.  If a pending
#        changelist number is given with the -c flag the opened file is
#        associated with that changelist, otherwise it is associated with
#        the 'default' pending changelist.
#
#        Files that are deleted generally do not appear on the have list.
    def _do_remove(self, files, force, recursive, p4=None):
        _files = self._splitFiles(files)
        output = ''
        if p4 is None:
            p4 = self.p4
        for basedir in _files.keys():
            result, raw = p4.delete(_files[basedir], _raw=1, cwd=basedir)
            raw_stderr = raw.get('stderr')
            if raw_stderr:
                raise p4lib.P4LibError(raw_stderr)
            if 'stdout' in raw:
                output += raw['stdout']

        forceRefresh = False
        self._fileStatusSvc.updateStatusForUris(files, forceRefresh)
        return output

    def remove(self, files, force, recursive, async_callback):
        return runP4CommandAsynchronously("remove", self, self._do_remove,
                                          files, True, async_callback, files,
                                          force, recursive)
    
#    sync -- Synchronize the client with its view of the depot
#
#    p4 sync [ -f -n ] [ file[revRange] ... ]
#
#        Sync updates the client workspace to reflect its current view (if
#        it has changed) and the current contents of the depot (if it has
#        changed).  The client view is used to map client file names to
#        depot file names and vice versa.
#
#        Sync adds files that are in the client view but which have not been
#        retrieved before.  Sync deletes previously retrieved files which
#        are no longer in the client view or have been deleted from the
#        depot.  Sync updates files which are still in the client view and
#        which have been updated in the depot.
#
#        Normally, sync affects all files in the client workspace.  If file
#        arguments are given, sync limits its operation to those files.
#        The file arguments may contain wildcards.
#
#        If the file argument includes a revision specifier, then the given
#        revision is retrieved.  Normally, the head revision is retrieved.
#        See 'p4 help revisions' for help specifying revisions.
#
#        If the file argument includes a revision range specification, then
#        only files selected by the revision range are updated, and the
#        highest revision in the range is used.
#
#        Normally, sync will not clobber files in the client workspace that
#        the user has made writable.  Setting the 'clobber' option in the
#        client spec disables this safety check.
#
#        The -f flag forces resynchronization even if the client already
#        has the file, and clobbers writable files.  This flag doesn't affect
#        open files.
    def _do_update(self, files, options, p4=None):
        _files = self._splitFiles(files)
        output = ''
        if p4 is None:
            p4 = self.p4
        for basedir in _files.keys():
            result, raw = p4.sync(_files[basedir], _raw=1, cwd=basedir)
            if 'stderr' in raw:
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
            #XXX need to deal with this
            ## build a better status output for komodo
            #if 'files' in result:
            #    for file in result['files']:
            #        localFileName = os.path.normpath(os.path.join(basedir,file['file']))
            #        url = uriparse.localPathToURI(localFileName)
            #        #print "%s %s %s" %(file['status'],file['file'],url)
            #        output += "%s %s %s\n" % (file['status'],file['file'],url)
            #output = ""
                    
        forceRefresh = False
        self._fileStatusSvc.updateStatusForUris(files, forceRefresh)
        return output

    def update(self, files, options, async_callback):
        return runP4CommandAsynchronously("update", self, self._do_update,
                                          files, False, async_callback, files,
                                          options)

#    revert -- Discard changes from an opened file
#
#    p4 revert [ -a -c changelist# ] file ...
#
#        Revert an open file back to the revision previously synced from
#        the depot, discarding any pending changelists or integrations that
#        have been made.  This command requires naming files explicitly.
#        After running revert the named files will no longer be locked
#        or open.
#
#        The -a flag tells 'p4 revert' to revert only those files which
#        are opened for edit or integrate and are unchanged or missing.
#        Files with pending integration records are left open.  With the
#        -a flag, the file arguments are optional.
#
#        The -c flag limits 'p4 revert' to files opened under the given,
#        pending changelist.
    def _do_revert(self, files, options, p4=None):
        # XXX unfortunately this function requires knowledge of
        # item properties
        _files = self._splitFiles(files)
        if p4 is None:
            p4 = self.p4
        output = ''
        for basedir in _files.keys():
            result, raw = p4.revert(_files[basedir], _raw=1, cwd=basedir)
            if 'stderr' in raw:
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']
            
        forceRefresh = False
        self._fileStatusSvc.updateStatusForUris(files, forceRefresh)
        return output

    def revert(self, files, options, async_callback):
        return runP4CommandAsynchronously("revert", self, self._do_revert,
                                          files, True, async_callback, files,
                                          options)

    def _do_edit(self, files, p4=None):
        # XXX unfortunately this function requires knowledge of
        # item properties
        _files = self._splitFiles(files)
        if p4 is None:
            p4 = self.p4
        output = ''
        for basedir in _files.keys():
            result, raw = self.p4.edit(_files[basedir], _raw=1, cwd=basedir)
            if 'stderr' in raw:
                output += raw['stderr']
            if 'stdout' in raw:
                output += raw['stdout']

        forceRefresh = False
        self._fileStatusSvc.updateStatusForUris(files, forceRefresh)
        return output

    def edit(self, files, async_callback):
        return runP4CommandAsynchronously("edit", self, self._do_edit,
                                          files, True, async_callback, files)


class P4AsyncOp(koAsyncOperationBase):

    def stop(self):
        # p4 is a p4lib.P4 instance
        p4 = self.kwargs.get('p4')
        if p4:
            self.status = components.interfaces.koIAsyncOperation.STATUS_STOPPING
            # XXX - need to implement p4.stop() to stop the process(es)
        else:
            koAsyncOperationBase.stop(self)

def runP4CommandAsynchronously(cmd_name, p4Svc, run_function,
                               associated_uris, lock_these_uris,async_callback,
                               *args, **kwargs):
    # Run asynchronously
    async_svc = components.classes["@activestate.com/koAsyncService;1"].\
                    getService(components.interfaces.koIAsyncService)
    # Create a new p4 instance to ensure we don't have multiple scc
    # operations running on the same instance.
    p4lib_instance = p4lib.P4()
    # Ensure the p4lib instance uses the same executable as the p4 service.
    p4lib_instance.p4 = p4Svc.p4.p4
    kwargs['p4'] = p4lib_instance
    async_op = P4AsyncOp(run_function, *args, **kwargs)
    async_svc.run("SCC p4 " + cmd_name, async_op, async_callback,
                  associated_uris, lock_these_uris)
    return async_op


from fileStatusUtils import KoSCCChecker

class KoP4FileChecker(KoSCCChecker):
    _reg_clsid_ = "{31900ece-3bb6-4973-ad70-8b0f4ba36902}"
    _reg_contractid_ = "@activestate.com/koFileStatusChecker?type=p4;1"
    _reg_desc_ = "Komodo Perforce File Status Checker"
    _reg_categories_ = [
         ("category-komodo-file-status",      "p4"),
         ]

    def __init__(self):
        KoSCCChecker.__init__(self, 'p4')
        #self.log.setLevel(logging.DEBUG)
        self.p4 = p4lib.P4()
        self.enabledPrefName = 'p4Enabled'
        self.executablePrefName = 'p4Executable'
        self.backgroundEnabledPrefName = 'p4BackgroundCheck'
        self.backgroundDurationPrefName = 'p4BackgroundMinutes'
        self.recursivePrefName = 'p4Recursive'
        self.setExecutable(self.svc.executable)

    # Overriding parent setExecutable
    def setExecutable(self, executable):
        if self._is_windows:
            self.executable = "p4.exe"
        else:
            self.executable = "p4"
        if executable and os.path.exists(executable):
            self.executable = executable

    def _cacheSCCInfo(self, cache_key, path, sccInfo):
        koSccInfo = {}
        onDisk = os.path.exists(path)
        sccInfo['ondisk'] = onDisk

        koSccInfo['sccType'] =  self.name
        koSccInfo['sccDirType'] = self.name
        koSccInfo['sccHaveOnDisk']= onDisk
        koSccInfo['sccLocalRevision'] = str(sccInfo['haveRev'])
        koSccInfo['sccDepotRevision'] = str(sccInfo['headRev'])
        koSccInfo['sccNeedSync'] = int(koSccInfo['sccLocalRevision'] != koSccInfo['sccDepotRevision'])
        koSccInfo['sccConflict'] = 0
        # necessary for status icons
        if koSccInfo['sccNeedSync']:
            koSccInfo['sccStatus'] = 'out of date'
            koSccInfo['sccSync'] = 1
        else:
            koSccInfo['sccStatus'] = 'up to date'
            koSccInfo['sccOk'] = 1
        koSccInfo['sccAction'] = sccInfo['action']
        koSccInfo['sccChange'] = sccInfo['change']

        self._cached_info[cache_key] = koSccInfo

    def updateSCCInfo(self, dir_nsUri, reason):
        # we use nbasepath for the directory key and basepath for running cmds
        self.p4.p4 = self.executable
        # execute p4 on the path
        if self.recursive:
            subpath = "..."
        else:
            subpath = "*"
        try:
            # We don't have any cached info and we haven't check this path yet,
            # so we do that now
            #print "p4 checking ",basepath + subpath
            path = dir_nsUri.path
            p4fstat = self.p4.fstat(subpath, cwd=path)
        except p4lib.P4LibError, e:
            # Don't notify errors for paths that are not under SCC
            if str(e).find("is not under client's root") == -1:
                self.notifyError('Perforce Status Error, see SCC Output pane for details', e)
            return False

        if not p4fstat:
            self.log.debug("No p4 info returned")
            return False

        # Cache all the file scc information
        tmp_nsUri = dir_nsUri.clone()
        dir_cache_key = self._norm_uri_cache_key(dir_nsUri.spec)
        for fileinfo in p4fstat:
            # Path is given as the full local file path:
            #   C:\\myrepo\\path\\file.txt
            #   /myrepo/path/file.txt
            fpath = fileinfo['path']
            if self._is_windows:
                tmp_nsUri.path = fpath.replace("\\", "/")
            else:
                tmp_nsUri.path = fpath
            cache_key = self._norm_uri_cache_key(tmp_nsUri.spec)
            if self.recursive:
                # Check the paths, we may need to update for a different dir
                parent_cache_key = os.path.dirname(cache_key)
                # Ensure it's compatible with nsURI.
                parent_cache_key += "/"
                if parent_cache_key != dir_cache_key:
                    self._lastChecked[parent_cache_key] = time.time()
            self._cacheSCCInfo(cache_key, fpath, fileinfo)

        return True
