#!python
# Copyright (c) 2000-2008 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

# The implementation of the Komodo Sleuth service

import os
from os.path import join, isabs, expanduser, normpath
import sys
import string, types
import re
import logging
import threading
import pprint
import which

from xpcom import components, nsError, ServerException, COMException
from xpcom.server import UnwrapObject
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject

import sleuth



#---- globals

log = logging.getLogger("koSleuth")
#log.setLevel(logging.DEBUG)



#---- Sleuth backend

class _SleuthRunner(threading.Thread):
    """Launch a unittest run based on the provided directory, language,
    and, optionally, command-line.  Feed results into the koISleuthResultsTreeView
    object.
    
    The running thread can be terminated early by calling the .stop() method.

    """
    
    def __init__(self, cwd, language, commandLine, resultsMgr, resultsView):
        threading.Thread.__init__(self, name="SleuthRunner")

        self.cwd = cwd
        self.language = language
        self.commandLine = commandLine
        self.resultsMgrProxy = getProxyForObject(None,
            components.interfaces.koISleuthResultsTabManager,
            resultsMgr, PROXY_ALWAYS | PROXY_SYNC)
        self.resultsViewProxy = getProxyForObject(None,
            components.interfaces.koISleuthResultsView,
            resultsView, PROXY_ALWAYS | PROXY_SYNC)
        self._resultsView = UnwrapObject(resultsView) # This might not be used.
        
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                      .getService(components.interfaces.koILastErrorService)

        self._is_running = False # because we allow only one testrun at a time
        self._event_dispatch_table = {
            sleuth.SuiteStartEvent : self._handleSuiteStartEvent,
            sleuth.SuiteFinishedEvent : self._handleSuiteFinishedEvent,
            sleuth.TestEvent : self._handleTestEvent,
            sleuth.TestFaultEvent : self._handleTestFaultEvent,
            sleuth.TestFinishedEvent : self._handleTestFinishedEvent,
            sleuth.TestInfoEvent : self._handleTestInfoEvent,
            sleuth.TestStartEvent : self._handleTestStartEvent,
        }
        self._inSuite = False
        self._cachedData = []
        self._currentDataItem = None
        self._cachedDataThreshold = 1
        self._firstInterestingItem = -1
        self._notifiedInterestingItem = False
        self._numItemsSent = 0
        
        # Use the user's python, and no need to change directories
        self.pythonDefaultInterpreter = None
        if self.language == "Python":
            pythonInfoEx = components.classes["@activestate.com/koAppInfoEx?app=Python;1"].\
                           createInstance(components.interfaces.koIAppInfoEx)
            self.pythonDefaultInterpreter = pythonInfoEx.executablePath

    def stop(self):
        """Stop processing."""
        log.info("stopping test_plan thread")
        self.test_plan.stop()
            
    def run(self):
        try:
            #self.resultsMgrProxy.runStarted()
            return self._run_in_ui_mode()
        finally:
            self.resultsMgrProxy.runFinished()
        
    def _run_in_ui_mode(self):
        log.debug("start sleuth thread")
        self._num_tests = self._num_passes = self._num_errors = self._num_failures = 0

        pythonExecutable = None
        if self.language == "Python":
            # Use the user's python, and no need to change directories
            launchDir = self.cwd
            pythonExecutable = self.pythonDefaultInterpreter
            usingKomodoPython = False

        if pythonExecutable is None:
            koDirs = components.classes["@activestate.com/koDirs;1"].\
                       getService(components.interfaces.koIDirs)
            pythonExecutable = koDirs.pythonExe
            launchDir = koDirs.mozBinDir
            usingKomodoPython = True  # As opposed to the user's installed Python

        self.test_plan = sleuth.TestPlan(self.cwd, self.language, self.commandLine, pythonExecutable, launchDir, usingKomodoPython)
        log.debug("In koSleuth.run, self.cwd=%s, self.language=%s, self.commandLine=[%s]",
                  self.cwd, self.language, self.commandLine)
        try:
            for event in self.test_plan.run_tests():
                log.debug("Got event %r", event)
                (self._event_dispatch_table.get(
                    event.__class__, self._handleUnknownEvent))(event)
        except StopIteration:
            log.debug("Done running test suite")
            self._relayDataIfReady(True)
        
    def _handleUnknownEvent(self, event):
        log.debug("Got unknown event %r", event)
        
    def _handleCurrentDataItem(self, force_relay=False):
        if self._currentDataItem:
            if self._firstInterestingItem == -1 and self._currentDataItem[2]:
                self._firstInterestingItem = len(self._cachedData)
            self._cachedData.append(self._currentDataItem)
            self._relayDataIfReady(force_relay)
        self._currentDataItem = None            

    def _relayDataIfReady(self, force_relay):
        if force_relay or len(self._cachedData) >= self._cachedDataThreshold:
            # Send each row to the tree view object
            # one at a time.
            for item in self._cachedData:
                log.debug("About to addDetails: %s",
                          ", ".join([repr(x) for x in item]))
                self.resultsViewProxy.AddDetails(item[0], item[1], "\n".join(item[2]))
            self.resultsViewProxy.UpdateTree()
            if self._firstInterestingItem > -1 and not self._notifiedInterestingItem:
                self._notifiedInterestingItem = True
                self.resultsMgrProxy.highlightLine(self._numItemsSent + self._firstInterestingItem)
                
            self._numItemsSent += len(self._cachedData)
            self._cachedData = []
        
    def _handleSuiteStartEvent(self, event):
        if event.data:
            self._handleCurrentDataItem()
            self._currentDataItem = ["Information", "", [event.data]]
        self._inSuite = True

    def _handleSuiteFinishedEvent(self, event):
        if not self._inSuite:
            return
        status_string = event.get_status_string()
        log.debug("_handleSuiteFinishedEvent: status=%r, time=%r",
                  status_string, event.elapsed_time)
        self._handleCurrentDataItem(True)
        # And have the UI display part
        vals = event.get_status_values()
        self._num_tests += vals[0]
        self._num_passes += vals[1]
        self._num_errors += vals[2]
        self._num_failures += vals[3]
        elapsed_time = event.elapsed_time
        if elapsed_time:
            elapsed_time = re.sub(r'^\s*T:\s*', '', elapsed_time)
        self.resultsMgrProxy.updateResults(self._num_passes,
                                           self._num_errors,
                                           self._num_failures,
                                           elapsed_time)
        self._inSuite = False
        
        
    def _handleTestFaultEvent(self, event):
        if event.info:
            if self._currentDataItem:
                self._currentDataItem[2] += event.info
            else:
                self._currentDataItem = ["Information", "", event.info]

    def _handleTestFinishedEvent(self, event):
        self._currentDataItem[1] = event.outcome
        self._handleCurrentDataItem()

    def _handleTestStartEvent(self, event):
        self._handleCurrentDataItem()
        self._currentDataItem = [event.name, "", []]
    
    def _handleTestInfoEvent(self, event):
        if event.info:
            if self._currentDataItem:
                self._currentDataItem[2] += event.info
            else:
                self._currentDataItem = ["Information", "", event.info]
        
    def _handleTestEvent(self, event):
        # Shouldn't happen
        log.debug("**** What do I do with a %r", event)
        if self._currentDataItem:
            self._currentDataItem[2] += event.info


class KoSleuthService:
    _com_interfaces_ = components.interfaces.koISleuthService
    _reg_desc_  = "Sleuth UnitTest Service"
    _reg_clsid_ = "{ca5ccaff-35d7-4785-9ee3-a98055b59659}"
    _reg_contractid_ = "@activestate.com/koSleuthService;1"

    def __init__(self):
        global lastErrorSvc
        lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                       .getService(components.interfaces.koILastErrorService)
        self._sleuthThread = None
       
    def run_sleuth_tests(self, cwd, language, commandLine,
                         resultsMgr, resultsView):
        """
        Run Sleuth unit-tests, grab results, feed into the given koISleuthResultsView.

            "cwd" : [string] directory to run tests in
            "language" : [string]
            "commandLine" : [string] optional command to run, instead of defaults
            "resultsMgr" is a koISleuthResultsTab instance.
            "resultsView" is a koISleuthResultsView instance
        
        This process is done asynchronously -- i.e. a separate thread is
        started to do this.
        
        No return value.
        """
        log.debug("run_sleuth_tests(cwd='%s', language=%s, commandLine=%r)",
                 cwd, language, commandLine)
        self._sleuthThread = _SleuthRunner(cwd, language, commandLine,
                                           resultsMgr, resultsView)
        resultsMgr.runStarted()
        self._sleuthThread.start()

    def stop_sleuth_tests(self):
        #XXX Do I need a lock-guard around usage of self._sleuthThread?
        if self._sleuthThread:
            self._sleuthThread.stop()
            self._sleuthThread = None
            
    def thread_ended(self):
        """
        When the UI gets a test-run-finished event, it can update the back-end
        """
        #XXX Grab mutex
        try:
            if self._sleuthThread:
                self._sleuthThread = None
        finally:
            #XXX Release mutex
            pass
            

