#!python

import os
from os.path import join, exists, dirname, isdir
import sys
import re
from pprint import pprint
import process
import subprocess

try:
    import which
    haveWhich = True
except ImportError:
    haveWhich = False

CREATE_NO_WINDOW = 0x8000000 # same as win32process.CREATE_NO_WINDOW
CREATE_DEFAULT_ERROR_MODE = 0x04000000

_VAGUE_PLAN = -1
""" Patterns, and what we're looking for:

Put all patterns here once we know what we're looking for.
"""
# Assume test names never contain two contiguous dots
#TODO: Refine this if the test name ends with \.t\b
_test_name_re = re.compile(r'(t[\\/].+?)\.{2,}\s*(.*)\s*$')

_simple_path_re = re.compile(r'[a-zA-Z0-9_\\.:]+$')
# Banners and other patterns at the start that we can safely ignore
_start_patterns_to_ignore = [re.compile(s) for s in (
                            r'perl.*blib.lib.*blib.arch',
                            r'Microsoft.*Program Maintenance Utility',
                            r'Copyright.*Microsoft Corp.*All rights reserved',
                            # vc6 - Copyright (C) Microsoft Corp 1988-1998. All rights reserved.
                            # vc7 - Copyright (C) Microsoft Corporation.  All rights reserved.
                            # vc8 - Copyright (C) Microsoft Corporation.  All rights reserved.
                            r'make.*Entering directory',
                            r'make.*Leaving directory',
                            r'^cp lib.*blib',
                            r'^\t', # Ignore as a make command line
                         )]

_died_failed_tests_re = re.compile(r"DIED\.\s+FAILED tests? \d")
_failed_summary_re = re.compile(r'Failed \d+/\d+ test (?:script|program)s?')
_failed_test_list_re = re.compile(r"FAILED tests? \d")
_failed_cant_re = re.compile(r"failed:\s*Can't")
_failed_test_stat_summary_re = re.compile(r'''Failed\s+Test\s+Stat\s+
                                            (?:Wstat\s+)?
                                            (?:Total\s+)?
                                            (?:Fail\s+)?
                                            (?:List\s+of\s+Failed)?''', re.X)
_failed_num_re = re.compile(r'\s*\#\s*Failed')

_files_tests_summary_re = re.compile('Files=\d+, Tests=\d+,')
_test_got_something_re = re.compile(r'# Test \d+ got:.*\bat line \d')

_not_ok_line_re = re.compile(r'not ok\s+(\d+)\s*(.*)\s*')
_ok_line_re = re.compile(r'ok\s+(\d+)\s*\#\s*skip(?:ped|ping)?\s*(.*)\s*')
_ok_line_simple_re = re.compile(r'ok\s+(\d+)\s*(.*)\s*')

_num_dot_dot_num_re = re.compile(r'(\d+)\s*\.\.\s*(\d+)\s*$')
_undefined_sub_re = re.compile(r'Undefined subroutine .*? called at')

class KoHarnessPerlTests(object):
    """If there's a Makefile, a test.pl, or a t/ directory, simply run the
    tests through Python's subprocess module to fold stderr and stdout, and
    convert the results to Komodo test output format.
    
    See unittest/koHarnessRuby.py for a description of the format
    this file should emit.
    """
    def __init__(self, cwd, cmd=None):
        self._cwd = cwd
        self._argv = cmd

    def set_perl(self, perl):
        self._perl = perl

    def final_initialize(self):
        if not self._argv:
            self._argv = self._get_argv()

    def _get_argv(self, rebuild_makefile=True):
        # Determine the command to run
        if exists(join(self._cwd, "Makefile")):
            make_exe = None
            if sys.platform == "win32":
                make_exe = _find_nmake_exe()
            elif haveWhich:
                try:
                    make_exe = which.which("make")
                except which.WhichError:
            
                        pass
            else:
                make_exe = process.ProcessOpen(['which', 'make']).communicate()[0]
                if make_exe:
                    make_exe = make_exe.rstrip()
            if make_exe:
                return [make_exe, "test"]

        # Either no Makefile, or we couldn't find a "make/nmake" to call
        # it.
        if exists(join(self._cwd, "Makefile.PL")) and rebuild_makefile:
            res = process.ProcessOpen([self._perl, 'Makefile.PL'],
                                      cwd=self._cwd).communicate()
            # Fake a test-suite
            print "@suite_started@: rebuilding the Makefile"
            print "@fault@:"
            if res[0]:
                print res[0];
            if res[1]:
                print res[1]
            print "@suite_finished@:"
            return self._get_argv(False)
        if exists(join(self._cwd, "test.pl")):
            return [self._perl, "test.pl"]
        elif isdir(join(self._cwd, "t")):
            s = [self._perl, "-MExtUtils::Command::MM",
                    "-e", "\"test_harness(0, 'blib/lib', 'blib/arch')\"",
                    "t/*.t"]
            return s
    
    def _ignore_at_start(self, line):
        for p in _start_patterns_to_ignore:
            if p.search(line): return True
        return False
        
    def pre_state_handle_test_end(self, line=None):
        print "@suite_finished@:"
        if line is not None:
            print line
        self.got_failure_line = False
        
    def state_handle_test_end(self, line):
        # Skip most unrecognized input at the end
        if not self.got_failure_line and _failed_summary_re.match(line):
            self.got_failure_line = True
            print line
        elif _files_tests_summary_re.match(line):
            print line
        elif line.startswith('All tests successful'):
            print line
        return True, self.state_handle_test_end
    
    def _process_line_in_plan(self, line):
        m = _not_ok_line_re.match(line)
        if m:
            self._test_outcome = "F"
            return True, m.group(1), '_num_failed', m.group(2)
        m = _ok_line_re.match(line)
        if m:
            self._test_outcome = "S"
            return True, m.group(1), '_num_skipped', m.group(2)
        m = _ok_line_simple_re.match(line)
        if m:
            self._test_outcome = "P"
            return True, m.group(1), '_num_passed', m.group(2)
        return False, None, None, None
                    
    def _outcome_to_tag(self, outcome):
        return outcome == "P" and "info" or "fault"
    
    def check_end_t(self, filename):
        for prefix in ("", self._cwd):
            full_name = join(prefix, filename)
            if not exists(full_name) and exists(full_name + ".t"):
                return filename + ".t"
        return filename

    def print_test_started(self, name=None):
        #@@@@ OUTPUT
        if name is None:
            print "@test_started@:"
        else:
            print "@test_started@: %s" % (self.check_end_t(name),)
     
    def print_test_finished(self):
        print "@test_result@: %s" % (getattr(self, '_test_outcome', 'P'),)
    
    def pre_state_in_plan(self):
        self.finish_state = self.post_state_in_plan
        
    def print_fault_tag_if_needed(self):
        if self._test_outcome != "P":
            print "@fault@:"
        
    def state_in_plan(self, line):
        if line.startswith("#"):
            if not self._have_details:
                self.print_fault_tag_if_needed()
                self._have_details = True
            print line
        elif 'ok' in line:
            if self._have_details:
                self._have_details = False
                self.print_test_finished()
            elif self.inside_a_test:
                self.print_test_finished()
            have_match, test_name, result_attr, details = self._process_line_in_plan(line)
            if have_match:
                self.print_test_started(test_name)
                self.inside_a_test = True
                setattr(self, result_attr, getattr(self, result_attr) + 1)
                self._num_ran += 1
                if details:
                    self._have_details = True
                    self.print_fault_tag_if_needed()
                    print details
        else:
            self.unrecognized_warning(line)
        return True, self.state_in_plan # Stay in this state until the end
    
    def post_state_in_plan(self):
        # Finish up the plan
        #XXX Use _VAGUE_PLAN
        self.print_test_finished()
        print "@suite_finished@: N:%d, P:%d, F:%d, S:%d\n" % (
            self._num_ran, self._num_passed, self._num_failed, self._num_skipped)

    def pre_state_start_of_plan(self):
        self._num_passed = 0
        self._num_failed = 0
        self._num_skipped = 0
        self._num_ran = 0
        self._have_details = False
    # ==> state_in_plan
    def state_start_of_plan(self, line):
        if line.startswith('#'):
            return True, self.state_start_of_plan # No transition
        if 'ok' in line:
            have_match, test_name, result_attr, details = self._process_line_in_plan(line)
            if have_match:
                self.print_test_started(test_name)
                self.inside_a_test = True
                setattr(self, result_attr, getattr(self, result_attr) + 1)
                self._num_ran += 1
                if details:
                    self._have_details = True
                    self.print_fault_tag_if_needed()
                    print details + "\n"
                self.pre_state_in_plan()
                return True, self.state_in_plan
        # Ignore this line, give warning, stay in this state
        self.unrecognized_warning(line)
        return True, self.state_start_of_plan
    
    def pre_state_before_suite(self):
        self.line = None
    # ==> state_handle_test_end, state_start_of_plan, state_in_suite_do_test_start
    def state_before_suite(self, line):
        if self._ignore_at_start(line):
            return True, self.state_before_suite # stay
        m = _failed_test_stat_summary_re.match(line)
        if m:
            self.pre_state_handle_test_end()
            return True, self.state_handle_test_end
        if line.startswith("Makefile out-of-date with respect"):
            self.pre_state_handle_test_end()
            return True, self.state_handle_test_end
        m = _num_dot_dot_num_re.match(line)
        if m:
            try:
                self._expected_num_tests = int(m.group(2)) - int(m.group(1)) + 1
            except:
                self._expected_num_tests = _VAGUE_PLAN
            #@@@@ OUTPUT
            print "@suite_started@:"
            self.pre_state_start_of_plan()
            return True, self.state_start_of_plan
        
        m = _test_name_re.match(line) #  (t[\\/]\w+)\.{2,}(.*)
        if m:
            print "@suite_started@:"
            self.pre_state_in_suite(m)
            return True, self.state_in_suite
        self.unrecognized_warning(line)
        return True, self.state_before_suite
    
    def unrecognized_warning(self, line):
        print "**** Unrecognized input: " + line
        if getattr(self, 'num_unrecognized_warnings', None) is None:
            self.num_unrecognized_warnings = 1
        else:
            self.num_unrecognized_warnings += 1
            if self.num_unrecognized_warnings >= 20:
                print "**** Bailing out early..."
                sys.exit(1)
        
    def pre_state_in_suite(self, m):
        self.inside_a_test = True
        self.inside_a_fault_or_info = False
        self.print_test_started(m.group(1))
        tag = m.group(2)
        if not tag:
            self._test_outcome = ""
        elif tag == "ok":
            self._test_outcome = "P"
        elif (tag.startswith("# Failed ")
              or tag.startswith("Can't")
              or _failed_test_list_re.search(tag)
              or _died_failed_tests_re.search(tag)
              or _undefined_sub_re.search(tag)
              or _test_got_something_re.match(tag)
              or _failed_cant_re.search(tag)):
            print '@fault@:' + tag
            self._test_outcome = "F"
            self.inside_a_fault_or_info = True
        else:
            self._test_outcome = "P"
            print '@info@:' + tag
            self.inside_a_fault_or_info = True
            
    def finish_state_in_suite(self):
        if self.inside_a_test:
            self.inside_a_fault_or_info = False
            self.print_test_finished()
            
    def state_in_suite(self, line):
        if line.startswith('All tests successful'):
            self.finish_state_in_suite()
            self.pre_state_handle_test_end(line)
            return True, self.state_handle_test_end
        m = _failed_test_stat_summary_re.match(line)
        if m:
            self.finish_state_in_suite()
            self.pre_state_handle_test_end()
            return True, self.state_handle_test_end
        # Assume test names end with ".t"
        m = _test_name_re.match(line) #  (t[\\/]\w+)\.{2,}(.*)
        if m:
            self.finish_state_in_suite()
            self.pre_state_in_suite(m)
            return True, self.state_in_suite
        
        if self.inside_a_fault_or_info:
            # Assume it's unrecognized, so emit it
            print line
        else:
            m = (_failed_num_re.match(line)
                 or _failed_test_list_re.search(line)
                 or _died_failed_tests_re.search(line))
            if m:
                print '@fault@:' + line
                self._test_outcome = "F"
                self.inside_a_fault_or_info = True
            else:
                self.unrecognized_warning(line)
        return True, self.state_in_suite # Always stay here

    def run_state_machine(self):
        while True:
            if self.line is None:
                line = self.p.stdout.readline()
                if not line:
                    break
                line = line.rstrip()
            else:
                line = self.line
                self.line = None
            if not line: continue
            
            # Go do the next state, maybe transition
            keep_running, next_state = self.next_state(line)
            if not keep_running or next_state is None:
                break
            if False and self.next_state != next_state:
                print "**** ==> state %r" % next_state
            self.next_state = next_state
    
    def init(self):
        self.curr_line = None

    def run(self):
        if not self._argv:
            # Report a failed test run in terms that sleuth will parse
            print "@suite_started@: failed attempt to run tests"
            print "@fault@:"
            print "No command-line specified, and no Makefile found in directory"
            print self._cwd
            print "@suite_finished@:"
            return
        if sys.platform.startswith("win"):
            # Don't display the Windows GPF dialog if the invoked program dies.
            # See comp.os.ms-windows.programmer.win32
            #  How to suppress crash notification dialog?, Jan 14,2004 - Raymond Chen's response
            # (URL http://groups.google.com/group/comp.os.ms-windows.programmer.win32/tree/browse_frm/thread/eeade5ecfa04d000/658b412a62ef5d7a?hl=en&rnum=1&q=suppressing+%22has+encountered+a+problem+and+needs+to+close%22+windows&_done=%2Fgroup%2Fcomp.os.ms-windows.programmer.win32%2Fbrowse_frm%2Fthread%2Feeade5ecfa04d000%2F74397b6bf9405304%3Fhl%3Den%26lnk%3Dst%26q%3Dsuppressing%2B%2522has%2Bencountered%2Ba%2Bproblem%2Band%2Bneeds%2Bto%2Bclose%2522%2Bwindows%26#doc_41fe71be34d3614b)
            import ctypes
            SEM_NOGPFAULTERRORBOX = 0x0002 # From MSDN
            ctypes.windll.kernel32.SetErrorMode(SEM_NOGPFAULTERRORBOX)
            flags = CREATE_NO_WINDOW
            shell = not self._argv[0].lower().endswith(".exe")
        else:
            flags = 0
            shell = False
        #print "#QQQ: koRunPerlTests.py: running cmd %s, dir %s" % (self._argv, self._cwd)
        self.p = subprocess.Popen(self._argv,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  shell=shell,
                                  cwd=self._cwd,
                                  # universal_newlines=True,
                                  creationflags=flags)
        #while True:
        #    d = self.p.stdout.readline()
        #    if not d: break
        #    print d,
        #self.p.wait()
        #return
        self.need_new_result_set = True
        self.init()
        self.pre_state_before_suite()
        self.next_state = self.state_before_suite
        self.run_state_machine()
        self.p.wait()



#---- internal support stuff

def _find_nmake_exe():
    if haveWhich:
        try:
            return which.which("nmake")
        except which.WhichError:
            pass
    return _find_nmake_in_registry()

def _find_nmake_in_registry():
    """Use well-known registry key names for versions 6, 7, and 8 of
    Visual Studio to determine where nmake.exe is located.
    
    The pattern on registry access used here is:
    
    childKey = _winreg.openKey(parentKey, string)
    try:
        use key childKey
    except EnvironmentError:
        pass # If the tryBlock does a _winreg.OpenKey(childKey, string)
    finally:
        _winreg.closeKey(childKey)
        
    It's a little different for cKey, because that might not be set.
    
    Permanent TODO: The list of supported versions in this docstring
    should be updated when each new version of VS is released.
    """
    import _winreg
    try:
        vskey = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                                r"SOFTWARE\Microsoft\VisualStudio")
        try:
            subkeyNames = []
            # Standard usage for enumerating a key is to try up to 1000 times.
            # See recipe 7.7 in Python Cookbook 1st ed.
            for vskey_count in range(1000):
                try:
                    subkeyNames.append(_winreg.EnumKey(vskey, vskey_count))
                except EnvironmentError:
                    break
            subkeyNames.sort(reverse=True) # Look at more recent first
            for sk in subkeyNames:
                subkey = _winreg.OpenKey(vskey, sk)
                try:
                    setupKey = _winreg.OpenKey(subkey, "Setup")
                    try:
                        try:
                            cKey = _winreg.OpenKey(setupKey, "VC")
                        except EnvironmentError:
                            try:
                                # VC++ 6 name
                                cKey = _winreg.OpenKey(setupKey, "Microsoft Visual C++")
                            except EnvironmentError:
                                continue
                        try:
                            productDir = _winreg.QueryValueEx(cKey, "ProductDir")[0]
                        except EnvironmentError:
                            # unlikely -- if we're at VS/x/Setup/VC|MSVC++, there's probably a ProductDir
                            continue
                        finally:
                            _winreg.CloseKey(cKey)
                        nmakePath = os.path.join(productDir, "bin", "nmake.exe")
                        if os.path.exists(nmakePath):
                            if _simple_path_re.match(nmakePath):
                                return nmakePath
                            else:
                                return nmakePath
                    finally:
                        _winreg.CloseKey(setupKey)
                except EnvironmentError:
                    # possible that $subkey/Setup doesn't exist
                    pass    
                finally:
                    _winreg.CloseKey(subkey)
        finally:
            _winreg.CloseKey(vskey)
    except EnvironmentError:
        # unlikely -- we're on a machine that has no VisualStudio installation
        pass

#---- mainline
        
if __name__ == "__main__":
    perl = None
    usage =  "usage: " + sys.argv[0] + "[ -x|--executable=interpreter-path] [cwd [command args...]]"
    import getopt
    optlist, args = getopt.getopt(sys.argv[1:], 'hx:', ['help', 'executable='])
    for opt, optarg in optlist:
        if opt in ('-x', '--executable'):
            perl = optarg
        elif opt in ('-h', '--help'):
            sys.stderr.write(usage + "\n")
            sys.exit()
    if len(args) == 0:
        cwd = os.getcwd()
        cmd = None
    else:
        cwd = args[0]
        os.chdir(cwd)
        if len(args) == 1:
            cmd = None
        else:
            cmd = args[1:]
    harness = KoHarnessPerlTests(cwd, cmd)
    harness.set_perl(perl)
    harness.final_initialize()
    harness.run()
