/* Copyright (c) 2006-2008 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

/**
 * Source Code Control history dialog.
 *
 * Usage:
 *  All dialog interaction is done via an object passed in and out as the first
 *  window argument: window.arguments[0].
 *      .urls           is the urls that have been selected by the user.
 *                      The URLs are automatically converted for display via
 *                      ko.uriparse.displayPath().
 *      .localRevisions is the revisions numbers for the given urls. The length
 *                      of the localRevisions array must match the length of
 *                      urls.
 *
 * How it works:
 *  Once the dialog is given the list of urls and has been opened, it will
 *  display the list of urls in a selectable tree. Once the user selects a
 *  url (can also happen on load), the history for that url will then be shown
 *  in the history tree. The history items can be selected and a diff can then
 *  be made on a particular revision or between two revisions.
 *  
 * Dev Notes:
 *  This dialog is similar in implementation to commit.xul (xbl me!).
 */

var gSCCHistoryDialog_log = ko.logging.getLogger("scc.history");
//gSCCHistoryDialog_log.setLevel(ko.logging.LOG_DEBUG);

var gSCCHistoryDialog = null;

function _sccHistoryTreeView(tree_type, initial_rows) {
    // Call the parent initializer.
    xtk.dataTreeView.apply(this, [initial_rows]);
    /* tree_type: 0 - urls, 1 - history items */
    this._history_tree_type = tree_type;
    /* localRevision is used to show which is the current version the user has */
    this._localRevisionHistoryItem = null;
};

_sccHistoryTreeView.prototype = new xtk.dataTreeView();
_sccHistoryTreeView.prototype.contructor = _sccHistoryTreeView;

_sccHistoryTreeView.prototype.getCellText = function(row, column)
{
    if (this._history_tree_type == 0) {
        return ko.uriparse.displayPath(this._rows[row]);
    } else {
        var koHistoryItem = this._rows[row];
        switch (column.id) {
            case 'scc_history_items_tree_column_revision':
                // Mark the current revision row with a " (your copy)" text
                if (koHistoryItem == this._localRevisionHistoryItem) {
                    return koHistoryItem.version + " (local revision)";
                }
                return koHistoryItem.version;
            case 'scc_history_items_tree_column_change':
                return koHistoryItem.change;
            case 'scc_history_items_tree_column_user':
                return koHistoryItem.author;
            case 'scc_history_items_tree_column_date':
                return koHistoryItem.date;
            case 'scc_history_items_tree_column_action':
                return koHistoryItem.action;
            case 'scc_history_items_tree_column_message':
                return koHistoryItem.message;
        }
        return "(Unknown column: " + column.id + ")";
    }
};

_sccHistoryTreeView.prototype.SortRevisionColumn = function(compare1, compare2)
{
    // Special sorter for revision values, as CVS revisions can have the form:
    //   10.2.45.111
    // Each compare argument is an object that contains these attributes:
    //   compare1.data      ->  data (string) return by getCellText.
    //   compare1.row       ->  the real underlying row object.
    //   compare1.oldindex  ->  original row index before the sort.
    if (compare1 && compare2) {
        if (compare1.data == compare2.data) {
            // Keep the sort order as uniform as possible.
            return compare1.oldindex - compare2.oldindex;
        } else {
            //return parseInt(compare1.data) - parseInt(compare2.data);
            var value1 = compare1.row.version.split(".");
            var value2 = compare2.row.version.split(".");
            for (var i=0; i < value1.length; i++) {
                if (i >= value2.length) {
                    //dump("Comparing " + value1 + " and " + value2 + "\n");
                    return 1;
                } else if (parseInt(value1[i]) > parseInt(value2[i])) {
                    return 1;
                } else if (parseInt(value1[i]) < parseInt(value2[i])) {
                    return -1;
                }
                /* else, they are the same and we continue looping */
            }
        }
    }
    return -1;
};

_sccHistoryTreeView.prototype.markRevisionAsLocal = function(revision) {
    //dump("Looking for revision: " + revision + "\n");
    this._localRevisionHistoryItem = null;
    // Find and select the row with this revision
    for (var i=0; i < this._rows.length; i++) {
        if (this._rows[i].version == revision) {
            this.selection.select(i);
            this._localRevisionHistoryItem = this._rows[i];
            //dump("Found local revision at row: " + i + "\n");
            break;
        }
    }
};


//---- interface routines for XUL

function scc_history_OnLoad()
{
    try {
        document.title = "SCC History";

        // .urls
        var urls = window.arguments[0].urls;
        if (typeof urls == "undefined" || urls == null || urls.length < 1) {
            //XXX Is this the kind of error handling we want to do in onload
            //    handlers?
            var msg = "Internal Error: illegal 'urls' value for "
                      +"SCC History dialog: '"+urls+"'.";
            gSCCHistoryDialog_log.error(msg);
            alert(msg);
            window.close();
        }

        var localRevisions = window.arguments[0].localRevisions;
        if (typeof localRevisions == "undefined" || localRevisions == null ||
            localRevisions.length != urls.length) {
            var msg = "Internal Error: illegal 'localRevisions' value for "
                      +"SCC History dialog: '"+localRevisions+"'.";
            gSCCHistoryDialog_log.error(msg);
            alert(msg);
            window.close();
        }

        // .sccHandler
        var scc_handler = window.arguments[0].sccHandler;
        if (typeof scc_handler == "undefined" || scc_handler == null) {
            //XXX Is this the kind of error handling we want to do in onload
            //    handlers?
            var msg = "Internal Error: illegal 'sccHandler' value for "
                      +"SCC History dialog: '" + scc_handler + "'.";
            gSCCHistoryDialog_log.error(msg);
            alert(msg);
            window.close();
        }
    
        gSCCHistoryDialog = new _scc_history_dialog(scc_handler, urls,
                                                    localRevisions);

        // Select the first url in the given list
        if (urls && urls.length > 0) {
            gSCCHistoryDialog.urlsTreeView.selection.select(0);
        }

        // Listen for keypress events (on capture)
        window.addEventListener("keypress", _scc_history_keypressHandler, true);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
        alert(e);
        window.close();
    }
}

function scc_history_OnUnload()
{
    window.removeEventListener("keypress", _scc_history_keypressHandler, true);
    if (gSCCHistoryDialog.async_op) {
        // Try and stop the asynchronous command from running.
        gSCCHistoryDialog.async_op.stop();
    }
}


function _scc_history_keypressHandler(event)
{
    gSCCHistoryDialog.handleKeypress(event);
}


function _scc_history_dialog(sccHandler, urls, localRevisions)
{
    try {
        // koISCC xpcom object
        this.sccHandler = sccHandler;
        // Array of strings
        this.urls = urls;
        // Array of strings, one for each url in urls
        this.localRevisions = localRevisions;
        // The local revision for the url currently selected
        this.localRevisionForUrl = "";
        /**
         * Asynchronous scc history operation running in the background.
         * @type Components.interfaces.koIAsyncOperation
         */
        this.async_op = null;
        // Last error service
        this._lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"]
                                .getService(Components.interfaces.koILastErrorService);
        // Prefs
        var prefsSvc = Components.classes["@activestate.com/koPrefService;1"].
                            getService(Components.interfaces.koIPrefService);
        this._prefs = prefsSvc.prefs;

        // Find some xul elements
        this.dialog = document.getElementById("dialog-scc-history");
        this.messageTextbox = document.getElementById("scc_history_message");
        this.urlsVbox = document.getElementById("scc_history_urls_vbox");
        this.urlsTree = document.getElementById("scc_history_urls_tree");
        this.historyTree = document.getElementById("scc_history_items_tree");
        this.history_revision_column = document.getElementById("scc_history_items_tree_column_revision");
        this.history_change_column = document.getElementById("scc_history_items_tree_column_change");
        this.history_action_column = document.getElementById("scc_history_items_tree_column_action");
        this.okButton = this.dialog.getButton("accept");
        this.cancelButton = this.dialog.getButton("cancel");
        this.cmd_diffToLocalVersion   = document.getElementById("cmd_sccHistory_diffToLocalVersion");
        this.cmd_diffToCurrentVersion = document.getElementById("cmd_sccHistory_diffToCurrentVersion");
        this.cmd_diffBetweenVersions  = document.getElementById("cmd_sccHistory_diffBetweenVersions");
        // Setup button access keys
        this.okButton.setAttribute("accesskey", "o");
        this.cancelButton.setAttribute("accesskey", "c");

        // show url section tree if there is more than one url
        if (this.urls.length > 1) {
            this.urlsVbox.setAttribute("collapsed", false);
            // get taller if there are a lot of urls
            if (this.urls.length > 20) {
                this.urlsTree.style.setProperty("min-height", "400px", "");
            }
        } else {
            this.urlsVbox.setAttribute("collapsed", true);
        }

        // Action is only applicable for perforce
        // XXX - This should come from the scc service object!
        if (this.sccHandler.name != "p4") {
            this.history_action_column.setAttribute("hidden", "true");
            this.history_change_column.setAttribute("hidden", "true");
        }

        // Setup the tree views
        this.urlsTreeView = new _sccHistoryTreeView(0, this.urls);
        this.urlsTree.treeBoxObject.view = this.urlsTreeView;

        this.historyTreeView = new _sccHistoryTreeView(1, []);
        this.historyTree.treeBoxObject.view = this.historyTreeView;

        // Size to content before moving so "moveToAlertPosition" calculations are
        // correct.
        window.sizeToContent();
        this.dialog.moveToAlertPosition();
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

// Load history info for a url
_scc_history_dialog.prototype.load_history_for_url = function (url)
{
    try {
        var options = "";
        // XXX - These options should come from the scc service!
        switch (this.sccHandler.name) {
            case 'p4':
                // XXX - Move these to a user controlled widget.
                //       These options should to be retrieved through the
                //       sccHandler.
                options = "L";    /* longFormat */
                options += "i";   /* follow branches */
                break;
        }
        var self = this;
        var log = gSCCHistoryDialog_log;
        var async_callback = {
            "callback": function(result, data) {
                try {
                    // Op is not longer running, don't need the reference to it.
                    self.async_op = null;
                    // data will be an array of koISCCHistoryItem,
                    // throw it into the tree view.
                    if (result == Components.interfaces.koIAsyncCallback.RESULT_SUCCESSFUL) {
                        var notifybox = document.getElementById("scc_history_notificationbox");
                        notifybox.removeAllNotifications(false /* false means to slide away */);
                        self.historyTreeView.setTreeRows(data);
                        // Select the first history item (most recent) in the list
                        if (data.length > 0) {
                            // Initially, sort the columns in descending
                            // revision order. For perforce, when listing
                            // branch info, the revision number gets restarted,
                            // so we sort perforce changes by the change number.
                            if (self.sccHandler.name == "p4") {
                                self.historyTreeView.sortDescending(self.history_change_column);
                            } else {
                                self.historyTreeView.sortDescending(self.history_revision_column);
                            }
                            self.historyTreeView.markRevisionAsLocal(self.localRevisionForUrl);
                            //this.historyTreeView.selection.select(0);
                        }
                        // Title: "filename : revision xyz"
                        document.title = ko.uriparse.baseName(url) + " : local revision is " + self.localRevisionForUrl;
                    } else if (result == Components.interfaces.koIAsyncCallback.RESULT_STOPPED) {
                        if (window)
                            window.close();
                    } else if (result == Components.interfaces.koIAsyncCallback.RESULT_ERROR) {
                        // data should contain the error message then.
                        if (data) {
                            ko.dialogs.alert(data);
                        } else {
                            ko.dialogs.alert("History could not be retrieved.");
                        }
                        if (window)
                            window.close();
                    }
                } catch (e) {
                    log.warn("could not load history results, exception: " + e);
                }
                self = null;
                log = null;
            }
        };
        
        var notificationbox = document.getElementById("scc_history_notificationbox");
        var label = "Fetching history information...";
        var value = null;
        var asyncSvc = Components.classes["@activestate.com/koAsyncService;1"].
                        getService(Components.interfaces.koIAsyncService);
        var image_src = asyncSvc.asynchronous_icon_url;
        notificationbox.appendNotification(label, value, image_src);
        this.async_op = this.sccHandler.history(url,  /* scc file   */
                                                options,
                                                async_callback);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

// Load history items for the given url
_scc_history_dialog.prototype.urls_onSelectionChange = function ()
{
    try {
        // Update the history items tree
        var uri = this.urls[this.urlsTree.currentIndex];
        this.localRevisionForUrl = this.localRevisions[this.urlsTree.currentIndex];
        this.load_history_for_url(uri);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

// Load history details for the given history item
_scc_history_dialog.prototype.history_onSelectionChange = function ()
{
    try {
        //dump("history_onSelectionChange\n");
        // Update the history message box
        if (this.historyTreeView.rows && this.historyTree.currentIndex >= 0) {
            //dump("this.historyTree.currentIndex: "  + this.historyTree.currentIndex + "\n");
            var message = this.historyTreeView.rows[this.historyTree.currentIndex].message;
            this.messageTextbox.value = message;
        }

        this.cmd_diffToLocalVersion.setAttribute("disabled",
                                                   (this.historyTreeView.selection.count != 1));
        this.cmd_diffToCurrentVersion.setAttribute("disabled",
                                                   (this.historyTreeView.selection.count != 1));
        this.cmd_diffBetweenVersions.setAttribute("disabled",
                                                   (this.historyTreeView.selection.count != 2));
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

_scc_history_dialog.prototype.performDiff = function (uri1, version1,
                                                      uri2, version2,
                                                      filepath, diffToLocal)
{
    try {
        // Get diff revision options
        // XXX - Should come from the scc service!
        var options = '';
        switch (this.sccHandler.name) {
            case 'p4':
                if (diffToLocal) {
                    if (this.historyTreeView._localRevisionHistoryItem &&
                        (this.historyTreeView._localRevisionHistoryItem.uri !=
                         uri1)) {
                        ko.dialogs.alert("Cannot perform diff between client file and a depot file on a different branch");
                        return;
                    }
                    uri1 = filepath + "#" + version1
                }
                break;
            case 'cvs':
                if (diffToLocal) {
                    options += " -r " + version1
                }
                break;
            case 'svn':
                options = this._prefs.getStringPref('svnDiffOptions');
                if (diffToLocal) {
                    options += " --revision " + version1
                }
                break;
            default:
                ko.dialogs.alert("Invalid Source Code Control Handler: "+this.sccHandler.name);
                window.setCursor("auto");
                return;
        }
        if (diffToLocal) {
            ko.scc.Diff(this.sccHandler, [uri1], options);
        } else {
            ko.scc.DiffRevisions(this.sccHandler,
                                 uri1, version1,
                                 uri2, version2,
                                 filepath, options);
        }

    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

_scc_history_dialog.prototype.diffToLocalVersion = function ()
{
    try {
        window.setCursor("wait");
        // Get the uri
        var filepath = this.urls[this.urlsTree.currentIndex];
        // Get the version number of the selected row
        var rowNum = this.historyTreeView.selection.currentIndex;
        var uri1 = this.historyTreeView.rows[rowNum].uri;
        var version1 = this.historyTreeView.rows[rowNum].version;

        this.performDiff(uri1, version1, '', '', filepath, true);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
    window.setCursor("auto");
}

_scc_history_dialog.prototype.diffToCurrentVersion = function ()
{
    try {
        window.setCursor("wait");
        // Get the uri
        var filepath = this.urls[this.urlsTree.currentIndex];
        // Get the version number of the selected row
        var rowNum = this.historyTreeView.selection.currentIndex;
        var uri1 = this.historyTreeView.rows[rowNum].uri;
        var version1 = this.historyTreeView.rows[rowNum].version;

        this.performDiff(uri1, version1, filepath, '', filepath, false);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
    window.setCursor("auto");
}

_scc_history_dialog.prototype.diffBetweenVersions = function ()
{
    try {
        // Get the uri
        window.setCursor("wait");

        // Get the selected tree rows
        var selected = [];
        var start = new Object();
        var end = new Object();
        var numRanges = this.historyTreeView.selection.getRangeCount();
        for (var t=0; t<numRanges; t++) {
            this.historyTreeView.selection.getRangeAt(t, start, end);
            for (var v = start.value; v <= end.value; v++){
                selected.push(v);
            }
        }
        //alert("Selected: " + selected);
        // Get the version numbers of the selected rows
        // XXX - For perforce, we need to take into account branches, as this
        //       means they may be a different filenames.
        var uri1 = this.historyTreeView.rows[selected[1]].uri;
        var version1 = this.historyTreeView.rows[selected[1]].version;
        var uri2 = this.historyTreeView.rows[selected[0]].uri;
        var version2 = this.historyTreeView.rows[selected[0]].version;
        var filepath = this.urls[this.urlsTree.currentIndex];

        this.performDiff(uri1, version1, uri2, version2, filepath, false);
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
    window.setCursor("auto");
}

_scc_history_dialog.prototype.doDefaultDiff = function ()
{
    try {
        if (this.historyTreeView.selection.count == 1) {
            this.diffToLocalVersion();
        } else if (this.historyTreeView.selection.count == 2) {
            this.diffBetweenVersions();
        }
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

_scc_history_dialog.prototype.handleKeypress = function (event)
{
    try {
        if (event.keyCode == event.DOM_VK_ENTER ||
            event.keyCode == event.DOM_VK_RETURN) {
            // We only want to fire this when we are not already focused on a
            // button, else it gets fired automatically by the button handler
            if (event.target.id != 'scc_history_diff_to_local_version_button' &&
                event.target.id != 'scc_history_diff_between_versions_button') {
                event.stopPropagation();
                event.preventDefault();
                event.cancelBubble = true;
                gSCCHistoryDialog.doDefaultDiff();
            }
        }
    } catch (e) {
        gSCCHistoryDialog_log.exception(e);
    }
}

